/*	encode.c
*/

#include "tik.h"

int
PNMtdci(char *filename)
{
	/* Returns number of frames processed */
	register uint32 PNMbytes;
	register uint8 *ref;
	register uint32 pos = 0;
	register uint32 same = 0;
	register int n;
	register double identical = 0;
	register double factored = 0;
	register int thresh = ((2.55 * e_percent) + 0.5);
	register tdciavg_t *avg = 0;
	register int wavexy = 0;
	double now = 0;
	off_t lseekpos = 0;
	uint32 improved = 0;

	/* Bound threshold to 0..255 */
	if (thresh < 1) thresh = 1;
	if (thresh > 255) thresh = 255;

	/* Check we have an image */
	if (PNMreader() == 0) return(0);
	if (wavetype != wavePNM) return(0);
	PNMbytes = wavec * (wavexy = (wavex * wavey));

	/* Create the file */
	if ((filename[0] == '-') && (filename[1] == 0)) {
		/* Use stdout */
		TDCIfile = stdout;
	} else {
		if (0 == (TDCIfile = fopen(filename, "wb"))) {
			ERROR(ERROR_CREATE, "could not create file %s\n", filename);
		}
	}

	/* Write TIK header */
	wavef = ((e_fps > 0) ?
		 ((uint32) ((1000000000.0 / e_fps) + 0.5)) :
		 ((uint32) ((1000000000.0 / 24) + 0.5)));
	e_fps = (1000000000.0 / wavef);
	wavet = ((e_t > 0) ?
		 ((uint32) ((e_t * 1000000000.0) + 0.5)) :
		 wavef);
	e_t = (wavet / 1000000000.0);
	waver = 0;
	waveg = ((uint32) ((1000000 * e_gamma) + 0.5));
	fprintf(TDCIfile,
"P%c\n"			// 6 for RGB or 5 for UYVYYY
"# TIK V %u %s\n"	// Version date, RGB or UYVYYY
"# TIK X %u\n"		// X dim in tuples (RGB or UYVYYY)
"# TIK Y %u\n"		// Y dim in pixels
"# TIK Z %u\n"		// Maximum value
"# TIK F %u\n"		// Frame time in nanseconds
"# TIK T %u\n"		// Shutter time in nanoseconds
"# TIK R %u\n"		// Rolling shutter time in nanoseconds
"# TIK G %u\n"		// Gamma * 1000000
"%d %d\n"		// X, Y in RGB or UYVYYY samples
"%d\n",			// Maximum value (same as Z)
		'6',
		VERSION,
		"RGB",
		PNMx,
		PNMy,
		255,
		wavef,
		wavet,
		waver,
		waveg,
		PNMx,
		PNMy,
		255);
	lseekpos = ftello(TDCIfile);

	/* Write initial image */
	fwrite(PNMimage, 1, PNMbytes, TDCIfile);

	/* If quality is set high enough, do averaging.
	   This picks the best byte value for each of the
	   factored pixel values, but that means it must
	   randomly update pixels already output...
	*/
	if (e_quality >= 90) {
		/* Allocate & initialize averaging buffer */
		register int pix;

		avg = calloc(wavexy, sizeof(tdciavg_t));
		for (pix=0; pix<wavexy; ++pix) {
			avg[pix].lseek = lseekpos;
			avg[pix].sum[0] = PNMimage[(3*pix)];
			avg[pix].sum[1] = PNMimage[(3*pix)+1];
			avg[pix].sum[2] = PNMimage[(3*pix)+2];
			avg[pix].count = 1;
			lseekpos += 3;
		}
	}
	lseekpos = ftello(TDCIfile);

	/* Write 0 spacer */
	fputc(0, TDCIfile);
	lseekpos = ftello(TDCIfile);

	/* Record differences from (updated) reference image */
	ref = PNMimage;
	for (n=1; n<e_n; ++n) {
		register int i;

		now += (1.0 / e_fps);
		INFO("Frame %d: %gs        \r", n, now);

		if (PNMreader() == 0) {
			e_n = n;
			goto done;
		}

		for (i=0; i<PNMbytes; i+=3) {
			/* Is any color channel value too unlikely to be
			   the same as the last value for this pixel?
			*/
			if ((PNMerr[ref[i]][PNMimage[i]][0] < thresh) ||
			    (PNMerr[ref[i+1]][PNMimage[i+1]][1] < thresh) ||
			    (PNMerr[ref[i+2]][PNMimage[i+2]][2] < thresh)) {
				/* Pixel RGB changed */
				do {
					/* Encode same as low first,
					   7 bits/byte, 0x80 continuation
					*/
					register int t = (same & 0x7f);
					if (same >>= 7) t |= 0x80;
					fputc(t, TDCIfile);
					++lseekpos;
				} while (same);

				/* Output new bytes */
				fwrite((PNMimage + i), 1, 3, TDCIfile);

				/* Handle averaging */
				if (avg) {
					register int pix = i / 3;

					/* Can only need fixing if count > 1 */
					if (avg[pix].count > 1) {
						uint8 v[3];
						v[0] = ((avg[pix].sum[0] / avg[pix].count) + 0.5);
						v[1] = ((avg[pix].sum[1] / avg[pix].count) + 0.5);
						v[2] = ((avg[pix].sum[2] / avg[pix].count) + 0.5);
						if ((ref[i] != v[0]) ||
						    (ref[1+1] != v[1]) ||
						    (ref[i+2] != v[2])) {
							/* Replace old value with better estimate */
							fseeko(TDCIfile, avg[pix].lseek, SEEK_SET);
							fwrite(&(v[0]), 3, sizeof(uint8), TDCIfile);
							fseeko(TDCIfile, lseekpos+3, SEEK_SET);
							++improved;
						}
					}

					avg[pix].sum[0] = PNMimage[i];
					avg[pix].sum[1] = PNMimage[i+1];
					avg[pix].sum[2] = PNMimage[i+2];
					avg[pix].count = 1;
					avg[pix].lseek = lseekpos;
				}
				lseekpos += 3;

				/* Update ref */
				ref[i] = PNMimage[i];
				ref[i+1] = PNMimage[i+1];
				ref[i+2] = PNMimage[i+2];
			} else {
				/* No change; bump same */
				++same;

				/* And update stats */
				++factored;
				identical += ((ref[i] == PNMimage[i]) &&
					      (ref[i+1] == PNMimage[i+1]) &&
					      (ref[i+2] == PNMimage[i+2]));

				/* Handle averaging */
				if (avg) {
					register int pix = i / 3;
					avg[pix].sum[0] += PNMimage[i];
					avg[pix].sum[1] += PNMimage[i+1];
					avg[pix].sum[2] += PNMimage[i+2];
					++(avg[pix].count);
				}
			}
		}

		/* Force "change" on very last pixel... */
		if ((same != 0) && (n == (e_n - 1))) {
			do {
				/* Encode same as low first,
				   7 bits/byte, 0x80 continuation
				*/
				register int t = (same & 0x7f);
				if (same >>= 7) t |= 0x80;
				fputc(t, TDCIfile);
				++lseekpos;
			} while (same);

			/* Output bytes */
			fwrite((PNMimage + PNMbytes - 3), 1, 3, TDCIfile);

			/* Handle averaging */
			if (avg) {
				register int pix = i / 3;

				/* Can only need fixing if count > 1 */
				if (avg[pix].count > 1) {
					uint8 v[3];
					v[0] = ((avg[pix].sum[0] / avg[pix].count) + 0.5);
					v[1] = ((avg[pix].sum[1] / avg[pix].count) + 0.5);
					v[2] = ((avg[pix].sum[2] / avg[pix].count) + 0.5);
					if ((ref[i] != v[0]) ||
					    (ref[1+1] != v[1]) ||
					    (ref[i+2] != v[2])) {
						/* Replace old value with better estimate */
						fseeko(TDCIfile, avg[pix].lseek, SEEK_SET);
						fwrite(&(v[0]), 3, sizeof(uint8), TDCIfile);
						fseeko(TDCIfile, lseekpos+3, SEEK_SET);
						++improved;
					}
				}

				avg[pix].sum[0] = PNMimage[i];
				avg[pix].sum[1] = PNMimage[i+1];
				avg[pix].sum[2] = PNMimage[i+2];
				avg[pix].count = 1;
				avg[pix].lseek = lseekpos;
			}
			lseekpos += 3;

		}

		free(PNMimage);
	}

done:
	INFO("%u frames @ %gs/frame, TDCI is %gs long\n",
	     n,
	     (1.0 / e_fps),
	     now);
	INFO("%0.0f pixels processed: %0.0f (%0.2f%%) identical, %0.0f (%0.2f%%) factored\n",
	     (n * (PNMbytes / 3.0)),
	     identical,
	     (100.0 * identical) / (n * (PNMbytes / 3.0)),
	     factored,
	     (100.0 * factored) / (n * (PNMbytes / 3.0)));
	if (improved) {
		INFO("%u pixels were improved by averaging factor values\n",
		     improved);
	}
	return(n);
}

