/*	noise.c
*/

#include "tik.h"

uint8	PNMerr[256][256][3];

int
PNMnoise(char *filename)
{
	/* Create an error model from PNM sequence
	   Error model is 256x256x3
	   Returns number of frames processed
	*/
	double noise[256][256][3];
	int PNMbytes;
	register uint8 *ref;
	register uint32 pos = 0;
	register uint32 same = 0;
	register int n, c, j, i;

	/* Clear error model */
	for (n=1; n<(3*256*256); ++n) *(&(noise[0][0][0]) + n) = 0;

	/* Check we have an image */
	if (PNMreader() == 0) return(0);
	PNMbytes = 3 * PNMx * PNMy;

	/* Record differences from (updated) reference image */
	INFO("Frame %d        \r", 0);
	ref = PNMimage;
	for (n=1; n<e_n; ++n) {
		register int i;

		if (PNMreader() == 0) {
			/* Ended early... use what we have */
			e_n = n;
		} else {
			UPDATE("Frame %d        \r", n);

			for (i=0; i<PNMbytes; i+=3) {
				++noise[ref[i]][PNMimage[i]][0];
				++noise[PNMimage[i]][ref[i]][0];
				++noise[ref[i+1]][PNMimage[i+1]][1];
				++noise[PNMimage[i+1]][ref[i+1]][1];
				++noise[ref[i+2]][PNMimage[i+2]][2];
				++noise[PNMimage[i+2]][ref[i+2]][2];
			}

			free(ref);
			ref = PNMimage;
		}
	}
	INFO("there were %d images found\n", e_n);
	free(ref);

	/* Normalize table to parts/255 */
	for (c=0; c<3; ++c) {
		for (j=0; j<256; ++j) {
			/* Get freq. of j followed by j,
			   but artificially make sure it happens
			   at least once to avoid divide by 0
			*/
			uint32 up[256], down[256];
			double jval;

			/* Ensure center value happens */
			++noise[j][j][c];

			/* Ensure monotonic approach/decay from center value */
			for (i=0; i<256; ++i) up[i] = (down[i] = noise[j][i][c]);
			for (i=0; i<j; ++i) if (up[i] > up[i+1]) up[i+1] = up[i];
			for (i=255; i>j; --i) if (up[i] > up[i-1]) up[i-1] = up[i];
			for (i=j; i>0; --i) if (down[i] < down[i-1]) down[i-1] = down[i];
			for (i=j; i<255; ++i) if (down[i] < down[i+1]) down[i+1] = down[i];

			/* Normalize to percent, rounding up */
#ifdef	LOG_ERR
			jval = log(1 + up[j] + down[j]);
#else
			jval = (1 + up[j] + down[j]);
#endif
			for (i=0; i<256; ++i) {
#ifdef	LOG_ERR
				double ival = log(1 + up[i] + down[i]);
#else
				double ival = (1 + up[i] + down[i]);
#endif

				PNMerr[j][i][c] = ((uint8) (((255.0 * ival) / jval) + 0.5));
			}
		}
	}

	/* Create the file */
	if ((filename[0] == '-') && (filename[1] == 0)) {
		/* Use stdout */
		TDCIfile = stdout;
	} else {
		if (0 == (TDCIfile = fopen(filename, "wb"))) {
			ERROR(ERROR_CREATE,
			      "could not create file %s\n",
			      filename);
		}
	}

	/* Noise map is really a 256x256 P6 image */
	fprintf(TDCIfile,
"P6\n"
"# TIK V %u NOISE\n"	// Version date, NOISE
"256 256\n"
"255\n",
		VERSION);
	fwrite(&(PNMerr[0][0][0]), 1, sizeof(PNMerr), TDCIfile);
	fflush(TDCIfile);

	return(n);
}
