/*	reader.c
*/

#include "tik.h"

int	PNMfd = -1;
uint8	PNMc;
uint32	PNMw;
int	PNMx = 0;
int	PNMy = 0;
int	PNMmaxval = 0;
uint8	*PNMimage = 0;
int	PNMtimedout = 0;

wave_t	waveref;
wave_t	wavework;
wave_t	*wavep = &wavework;

inline int
PNMnextc(void)
{
	/* Get next char into PNMc */
	PNMc = 0;
	if (read(PNMfd, &PNMc, 1) != 1) {
		PNMtimedout = 1;
		return(PNMc = 0);
	}

	return(PNMc);
}

inline uint32
PNMnextw(void)
{
	/* Get next 32-bit word into PNMw */

	*(((char *) &PNMw) + 0) = PNMc;
	*(((char *) &PNMw) + 1) = PNMnextc();
	*(((char *) &PNMw) + 2) = PNMnextc();
	*(((char *) &PNMw) + 3) = PNMnextc();
	PNMnextc();

	return(PNMw);
}

static	uint8	fastbuf[512];
static	int	fastsp = 0;
static	int	fasti = 0;

inline int
PNMnextcfast(void)
{
	/* Fast (buffered) get next char into PNMc */
	int size = 0;
	int n = 0;

	if (fasti < fastsp) {
		return(PNMc = fastbuf[fasti++]);
	}

	fasti = (fastsp = 0);
	if ((n = ioctl(PNMfd, FIONREAD, &size)) < 2) {
		return(PNMnextc());
	}

	if (n > 512) n = 512;
	if (read(PNMfd, &(fastbuf[0]), n) != n) {
		PNMtimedout = 1;
		return(PNMc = 0);
	}
	fastsp = n+1;
	fasti = 1;
	return(PNMc = fastbuf[0]);
}

inline uint8 *
PNMnextbn(uint8 *buf, int count)
{
	register uint8 *p = buf;

	/* Read of a large thing from a pipe can be
	   fragmented into pieces...
	*/
	while (count > 0) {
		register int t = read(PNMfd, p, count);
		if (t < 0) {
			return((uint8 *) 0);
		}
		count -= t;
		p += t;
	}

	return(buf);
}

static inline uint8 *
PNMnextn(int count)
{
	/* Get next count bytes into malloc'd space */
	register uint8 *p = malloc(count);
	register uint8 *q;

	if (p == 0) return((uint8 *) 0);

	if ((q = PNMnextbn(p, count)) == ((uint8 *) 0)) {
		free(p);
	}

	return(q);
}

static inline int
PNMisspace(char c)
{
	return((c == ' ') ||
	       (c == '\t') ||
	       (c == '\n') ||
	       (c == '\r'));
}

static inline void
PNMspace(void)
{
	/* Read past white space, comments */
	while (PNMisspace(PNMc) || (PNMc == '#')) {
		if (PNMc == '#') while (PNMnextc() != '\n');
		PNMnextc();
	}
}

static inline int
PNMnumber(void)
{
	/* Read & return a number, 0 for none */
	register int n;
	register int charval = PNMc;

	if ((charval < '0') || (charval > '9')) return(0);

	n = (charval - '0');
	charval = PNMnextc();
	while ((charval >= '0') && (charval <= '9')) {
		n *= 10;
		n += (charval - '0');
		charval = PNMnextc();
	}
	return(n);
}


static char	TIKwordbuf[1024];

static inline char *
TIKword(void)
{
	int off = 0;
	while ((PNMc == ' ') || (PNMc == '\t')) PNMnextc();
	while (!PNMisspace(PNMc)) {
		if (off < 1023) {
			TIKwordbuf[off++] = PNMc;
		}
		PNMnextc();
	}
	TIKwordbuf[off] = 0;
	return(&(TIKwordbuf[0]));
}

static inline void
TIKspace(void)
{
	/* Read past white space, comments */
	while (PNMisspace(PNMc) || (PNMc == '#')) {
		if (PNMc == '#') {
			while (PNMisspace(PNMnextc())) ;
			if ((PNMc == 'T') &&
			    (PNMnextc() == 'I') &&
			    (PNMnextc() == 'K') &&
			    (PNMnextc() == ' ')) {
				/* It's a TIK file extension */
				TIKword();
				if (TIKwordbuf[1] == 0) {
					char c = TIKwordbuf[0];

					TIKword();
					switch (c) {
					case 'V':
						wavev = atoi(&(TIKwordbuf[0]));
						TIKword();
						if (!strcmp(&(TIKwordbuf[0]), "GRAY")) {
							wavetype = waveGRAY;
							wavec = 1;
						} else if (!strcmp(&(TIKwordbuf[0]), "RGB")) {
							wavetype = waveRGB;
							wavec = 3;
						} else if (!strcmp(&(TIKwordbuf[0]), "UYVYYY")) {
							wavetype = ((wavev >= 20161130) ?
								    waveUYVYYY :
								    waveUYVYYYold);
							wavec = 6;
						} else if (!strcmp(&(TIKwordbuf[0]), "NOISE")) {
							wavetype = waveNOISE;
							wavex = 256;
							wavey = 256;
							wavez = 255;
							wavec = 3;
						}
						break;
					case 'X':
						wavex = atoi(&(TIKwordbuf[0]));
						break;
					case 'Y':
						wavey = atoi(&(TIKwordbuf[0]));
						break;
					case 'Z':
						wavez = atoi(&(TIKwordbuf[0]));
						break;
					case 'B':
						waveb = atoll(&(TIKwordbuf[0]));
						break;
					case 'F':
						wavef = atoi(&(TIKwordbuf[0]));
						break;
					case 'T':
						wavet = atoi(&(TIKwordbuf[0]));
						break;
					case 'R':
						waver = atoi(&(TIKwordbuf[0]));
						TIKword();
						waverx = atoi(&(TIKwordbuf[0]));
						TIKword();
						wavery = atoi(&(TIKwordbuf[0]));
						break;
					case 'G':
						waveg = atoi(&(TIKwordbuf[0]));
						break;
					}
				}
			}
			while (PNMc != '\n') PNMnextc();
		}
		PNMnextc();
	}
}

uint8 *
PNMreader()
{
	/* Read the PNM or PNM-disguised TIK file */
	int s;
	uint8 *p;

	/* Kludge because EOF doesn't go through a fifo */
	PNMtimedout = 0;
	PNMnextc();
	if (PNMtimedout) return(PNMimage = 0);

	/* Assume ordinary PNM until proven otherwise */
	wavetype = wavePNM;
	wavev = 0;
	wavex = 0;
	wavey = 0;
	wavez = 0;
	waveb = 0;
	wavef = 0;
	wavet = 0;
	waver = 0;
	waverx = 0;
	wavery = 0;
	waveg = 0;

	if (PNMc != 'P') goto notP6P5;
	switch (s = PNMnextc()) {
	case '6':	wavec = 3; break;
	case '5':	wavec = 1; break;
	default:
		/* We don't handle other types */
notP6P5:
		ERROR(ERROR_FORMAT,
		      "only P6 or P5 format PNMs can be processed\n");
	}
	PNMnextc();
	TIKspace();
	PNMx = PNMnumber();
	if (wavex == 0) wavex = PNMx;
	PNMspace();
	PNMy = PNMnumber();
	if (wavey == 0) wavey = PNMy;
	PNMspace();
	PNMmaxval = PNMnumber();
	if (wavez == 0) wavez = PNMmaxval;
	if (!PNMisspace(PNMc)) return(PNMimage = 0);

	/* Now copy in the (8*wavec)-bit data */
	return(PNMimage = PNMnextn(wavec * wavex * wavey));
}

uint8 *
P5reader()
{
	/* Read the P5 file */
	int x, y, z, s;
	uint8 *p;

	/* Kludge because EOF doesn't go through a fifo */
	PNMtimedout = 0;
	PNMnextc();
	if (PNMtimedout) return(PNMimage = 0);

	if (PNMc != 'P') goto notP6P5;
	switch (PNMnextc()) {
	case '5':	break;
	default:
		/* We don't handle other types */
notP6P5:
		ERROR(ERROR_FORMAT,
		      "only P5 format is allowed here\n");
	}
	PNMnextc();
	PNMspace();
	x = PNMnumber();
	PNMspace();
	y = PNMnumber();
	PNMspace();
	z = PNMnumber();
	if (!PNMisspace(PNMc)) return(PNMimage = 0);

	/* Now copy in the 8-bit data */
	return(PNMimage = PNMnextn(x * y));
}

uint8 *
P6reader()
{
	/* Read the P6 file */
	int x, y, z, s;
	uint8 *p;

	/* Kludge because EOF doesn't go through a fifo */
	PNMtimedout = 0;
	PNMnextc();
	if (PNMtimedout) return(PNMimage = 0);

	if (PNMc != 'P') goto notP6P5;
	switch (PNMnextc()) {
	case '6':	break;
	default:
		/* We don't handle other types */
notP6P5:
		ERROR(ERROR_FORMAT,
		      "only P6 format is allowed here\n");
	}
	PNMnextc();
	PNMspace();
	x = PNMnumber();
	PNMspace();
	y = PNMnumber();
	PNMspace();
	z = PNMnumber();
	if (!PNMisspace(PNMc)) return(PNMimage = 0);

	/* Now copy in the 24-bit data */
	return(PNMimage = PNMnextn(3 * x * y));
}

void
start_ffmpeg(char *filename)
{
	int child = -1;
	int pipefd[2];

	/* Make a pipe */
	if (pipe(pipefd) == -1) {
		ERROR(ERROR_FIFO,
		      "could not create pipe for ffmpeg\n");
	}

	/* Need ffmpeg process to feed us a PPM stream... */
	child = fork();
	if (child == 0) {
		/* Close read end of pipe, make write end stdout */
		close(pipefd[0]);
		dup2(pipefd[1], 1);
		close(pipefd[1]);

		/* Child runs ffmpeg, then terminates cleanly */
		char cmd[1024];
		register int e = 0;

		sprintf(cmd,
			"%s -i %s -f image2pipe -vcodec ppm - 2>/dev/null",
			FFMPEG,
			filename);
		e = system(cmd);
		exit(e);
	}

	/* Only the parent here... */
	PNMfd = pipefd[0];
	close(pipefd[1]);
}

void
stop_ffmpeg(void)
{
	/* Shut it down */
	close(PNMfd);
}
