/*	render.c
*/

#include "tik.h"

typedef struct {
	double	start, stop;	/* Exposure interval */
	double	qual;		/* Samples used to make image */
	double	divby;		/* For scaling */
	double	*image;
	int	more;		/* More pixels not yet done? */
} interval_t;

static	interval_t	render[MAXRENDER];
static	int		rendersp = 0;
static	char		*infile = 0;
static	int		dottik = 0;
static	double		minstart = 0;



#define	FOR_RENDER	for (r=0; r<rendersp; ++r)
#define	START		(render[r].start)
#define	STOP		(render[r].stop)
#define	QUAL		(render[r].qual)
#define	DIVBY		(render[r].divby)
#define	IMAGE		(render[r].image)
#define	MORE		(render[r].more)
#define	SUM(X)		*(IMAGE + (X))


static inline unsigned int
xinu(unsigned int x)
{
	unsigned int y;

	*(((char *) &y) + 0) = *(((char *) &x) + 3);
	*(((char *) &y) + 1) = *(((char *) &x) + 2);
	*(((char *) &y) + 2) = *(((char *) &x) + 1);
	*(((char *) &y) + 3) = *(((char *) &x) + 0);
	return(y);
}

static void
TIKrenderWriteFile(register int r)
{
	char outfile[4*1024];
	char *s;

	sprintf(outfile, e_outfile, r);
	INFO("writing \"%s\" as frame %d (%1.3fs @ %1.3fs) from \"%s\" with quality %1.3f\n",
	     outfile,
	     r,
	     (STOP-START) / 1000000000.0,
	     START / 1000000000.0,
	     infile,
	     QUAL);

	if ((s = tdci2jpeg(outfile, IMAGE)) != 0) {
		ERROR(ERROR_WRITE,
		      "%s %s\n",
		      outfile,
		      s);
	}

	MORE = 0;
	free(IMAGE);
	IMAGE = ((double *) 0);
}

void
TIKrenderSimul(void)
{
	register uint8 *p;
	register int r;

	/* Anything requested? */
	if (rendersp < 1) {
		ERROR(ERROR_EXPOSE,
		      "no exposures requested\n");
	}

	/* Open the input file */
	if (dottik) {
		if (0 > (PNMfd = open(infile, O_RDONLY))) {
			ERROR(ERROR_OPEN,
			      "could not open TIK file %s\n",
			      infile);
		}
	} else {
		start_ffmpeg(infile);
	}

	/* Initialize wave stuff */
	memcpy(&wavework, &waveref, sizeof(waveref));
	if (waveref.x > 0) {
		/* Non-zero X means this is next frame */
		wavework.b += waveref.f;
	}
	wavep = &wavework;

	/* Check we have an image */
	if (p = PNMreader()) {
		double gamma[256];
		register double invgamma;
		register uint32 xy;
		register uint32 xyc;
		register double *from;
		register uint8 *ref;
		register double wf;
		register double wt;
		register double now;
		register uint32 where;
		register int i, off, n, r;

		memcpy(&waveref, &wavework, sizeof(waveref));
		xy = wavex * wavey;
		xyc = xy * ((wavetype == wavePNM) ? 3 : wavec);
		from = calloc(xy, sizeof(double));

		/* Mark all frames as no pixels rendered */
		FOR_RENDER MORE = xy;

		ref = p; /* Set initial TDCI values */
		wf = ((wavetype == wavePNM) ? (1000000000.0 / 24) : wavef);
		wt = (((wavet < 1) || (wavet > wavef)) ? wavef : wavet);
		now = (waveb / 1000000000.0) + wf;
		n = 0;

		/* Create gamma mapping info */
		if (e_gamma == 0) {
			e_gamma = ((waveg == 0) ? 1.0 : (waveg / 1000000.0));
		}
		for (i=0; i<256; ++i) {
			gamma[i] = ((e_gamma == 1) ?
				    ((double) i) :
				    pow(((double) i), e_gamma));
		}
		invgamma = 1.0 / e_gamma;

		/* Stuff specific to different formats */
		switch (wavetype) {
		case waveRGB:
			/* Should be 0 separator; skip over it */
			if (PNMnextc()) goto badformat;
			PNMnextc();

			/* Process TDCI until a little past stop...
			   past stop because we need frame after so we can
			   interpolate the slope in the gap between samples
			   Initialize where to inc past end of first frame
			*/
			where = xy - 1;
			while (!PNMtimedout) {
				register uint32 shift = 0;
				register int c;
				uint8 v[3 + 1];	/* +1 to allow PNMnextbn hack below */
				register int natstart = n;
				register int left = 0;

				/* Skip to next changed pixel */
				++where;
				do {
					/* 7 bits of skip at a time, low first */
					where += (((c = PNMc) & 0x7f) << shift);
					shift += 7;
					while (where >= xy) {
						/* In another frame... */
						where -= xy;
						now += wf;
						++n;
						INFO("Frame %d: %1.3fs        \r",
						     n,
						     (now / 1000000000.0));
					}
					PNMnextc();
					if (PNMtimedout) {
badformat:
						ERROR(ERROR_FORMAT, "bad format in TIK RGB file\n");
					}
				} while (c & 0x80);

				/* Where are we? */
				off = where * wavec;

				/* Get value of this sample */
#ifdef	NOTNOW
				for (i=0; i<wavec; ++i) {
					v[i] = PNMc;
					PNMnextc();
					if (PNMtimedout) goto badformat;
				}
#else
				v[0] = PNMc;
				PNMnextbn(&(v[1]), wavec);
				if (PNMtimedout) goto badformat;
				PNMc = v[wavec];
#endif

				/* Add contribution from previous sample to sum */
				if (now > minstart) {
					FOR_RENDER {
						if ((MORE > 0) && (now > START)) {
							register double f = from[where];
							register double g = now - (wf - wt); /* start of gap between samples */
							register double fstart = ((START > f) ? START : f);
							register double fstop = ((g > STOP) ? STOP : g);
							register double gstart = ((START > g) ? START : g);
							register double gstop = ((now > STOP) ? STOP : now);
							register uint32 off = where * wavec;
		
							/* Newly rendering this frame? */
							if (IMAGE == ((double *) 0)) {
								if (!(IMAGE = calloc(rendersp, xyc * sizeof(double)))) {
									ERROR(ERROR_EXPOSE,
									      "cannot allocate memory for TIK virtual exposure");
								}
							}
		
							/* Add portion of previous sample */
							if (fstart < fstop) {
								register double w = fstop - fstart;
								for (i=0; i<wavec; ++i) SUM(off+i) += (gamma[ref[off+i]] * w);
		
								/* Add-in fraction from real data frames */
								QUAL += (1.0 + (n - natstart)) * (w / (now - f));
							}
		
							/* Add portion of gap as a slope */
							if (gstart < gstop) {
								register double wstart = (gstart - g) / (now - g);
								register double wstop = (gstop - g) / (now - g);
								register double wgap = gstop - gstart;
		
								for (i=0; i<wavec; ++i) {
									/* Compute gap end values, then average * wgap to get area */
									register double vstart = (ref[off+i] * (1.0 - wstart)) + (gamma[v[i]] * wstart);
									register double vstop = (ref[off+i] * (1.0 - wstop)) + (gamma[v[i]] * wstop);
									SUM(off+i) += (wgap * ((vstart + vstop) / 2.0));
								}
		
								/* Add-in gap fraction as if it is from one frame */
								QUAL += (wgap / (now - f));
							}
		
							/* Update more */
							if (now > STOP+wf) {
								/* Add-in data for other pixels */
								register int pos;
								register int off = 0;

								for (pos=0; pos<xy; ++pos) {
									register double f = from[pos];
									register double w = STOP - ((f < START) ? START : f);
									if (w < 0) w = 0;
									for (i=0; i<wavec; ++i) {
										SUM(off) = (SUM(off) + (gamma[ref[off]] * w)) * DIVBY;
										if (e_gamma != 1) {
											SUM(off) = pow(SUM(off), invgamma);
										}
										++off;
										++QUAL;
									}
								}

								/* Correct exposure quality */
								QUAL /= xy;

								/* Done with this virtual exposure; write the file */
								TIKrenderWriteFile(r);
							}
						}

						/* Still any images left incomplete? */
						left += (MORE > 0);
					}

					/* Check to see if any frames not yet rendered */
					if (left < 1) goto noneleft;
				}

				/* Record this sample */
				from[where] = now;
				for (i=0; i<wavec; ++i) ref[off+i] = v[i];
			}

			/* Update image with last samples...
			   which we'll assume persist until stop,
			   because we can't know otherwise  ;-)
			   Also divide by time to get average values.
			*/
			FOR_RENDER
			if (MORE > 0) {
				char outfile[4*1024];
				char *s;

				off = 0;
				for (where=0; where<xy; ++where) {
					register double f = from[where];
					register double w = STOP - ((f < START) ? START : f);

					if (w < 0) w = 0;
					for (i=0; i<wavec; ++i) {
						SUM(off) = (SUM(off) + (gamma[ref[off]] * w)) * DIVBY;
						if (e_gamma != 1) {
							SUM(off) = pow(SUM(off), invgamma);
						}
						++off;

						/* Add-in fraction from real data */
						++QUAL;
					}
				}

				/* Correct exposure quality */
				QUAL /= xy;

				/* Done with this virtual exposure; write the file */
				TIKrenderWriteFile(r);
			}
			break;

		case waveUYVYYYold:
			/* Should be a P5 header next...
			   we know how to read that!
			*/
			for (where=0; where<xyc; where+=6) {
				ref[where] += 128;
				ref[where+2] += 128;
			}
			p = P5reader();
			while (p) {
				register double f = now - wf;
				register double g = now - (wf - wt); /* start of gap between samples */
				register int left = 0;

				FOR_RENDER {
					if ((MORE > 0) && (now > START)) {
						register double fstart = ((START > f) ? START : f);
						register double fstop = ((g > STOP) ? STOP : g);
						register double gstart = ((START > g) ? START : g);
						register double gstop = ((now > STOP) ? STOP : now);
						register double wstart = (gstart - g) / (now - g);
						register double wstop = (gstop - g) / (now - g);
						register double wgap = gstop - gstart;

						/* Newly rendering this frame? */
						if (IMAGE == ((double *) 0)) {
							if (!(IMAGE = calloc(rendersp, xyc * sizeof(double)))) {
								ERROR(ERROR_EXPOSE,
								      "cannot allocate memory for UYVYYY virtual exposure");
							}
						}

						/* Update quality */
						++QUAL;

						for (where=0; where<xyc; where+=6) {
							p[where] += 128;
							p[where+2] += 128;
						}
						for (where=0; where<xyc; ++where) {
							/* Add contribution from previous sample to sum */
							if (now > START) {
								/* Add portion of previous sample */
								if (fstart < fstop) {
									register double w = fstop - fstart;
									SUM(where) += (gamma[ref[where]] * w);
								}

								/* Add portion of gap as a slope */
								if (gstart < gstop) {
									/* Compute gap end values, then average * wgap to get area */
									register double vstart = (ref[where] * (1.0 - wstart)) + (gamma[p[where]] * wstart);
									register double vstop = (ref[where] * (1.0 - wstop)) + (gamma[p[where]] * wstop);
									SUM(where) += (wgap * ((vstart + vstop) / 2.0));
								}
							}
						}

						/* Update more and write file */
						if (now >= STOP) {
							/* Done with this virtual exposure; write the file */
							TIKrenderWriteFile(r);
						}
					}

					/* Still any images left incomplete? */
					left += (MORE > 0);
				}

				/* Check to see if any frames not yet rendered */
				if (left < 1) goto noneleft;

				/* Next image */
				now += wf;
				free(ref);
				ref = p;
				p = P5reader();
			}

			/* Did we get another? If so, discard it. */
			if (p) free(p);

			/* Update image with last samples...
			   which we'll assume persist until stop,
			   because we can't know otherwise  ;-)
			   Also divide by time to get average values.
			*/
			FOR_RENDER
			if (MORE > 0) {
				/* Newly rendering this frame? */
				if (IMAGE == ((double *) 0)) {
					if (!(IMAGE = calloc(rendersp, xyc * sizeof(double)))) {
						ERROR(ERROR_EXPOSE,
						      "cannot allocate memory for UYVYYY virtual exposure");
					}
				}

				/* Update quality */
				++QUAL;

				for (where=0; where<xyc; ++where) {
					register double f = now - wf;
					register double w = STOP - ((f < START) ? START : f);

					if (w < 0) w = 0;
					SUM(where) = (SUM(where) + (gamma[ref[where]] * w)) * DIVBY;
					if (e_gamma != 1) {
						SUM(where) = pow(SUM(where), invgamma);
					}
				}

				/* Done with this virtual exposure; write the file */
				TIKrenderWriteFile(r);
			}
			break;

		case waveUYVYYY:
			/* Make all values signed */
			for (where=0; where<xyc; where+=6) {
				ref[where] += 128;
				ref[where+2] += 128;
			}

			/* Resize from[] */
			free(from);
			from = calloc(xyc, sizeof(double));

			/* Process TDCI until a little past stop...
			   past stop because we need frame after so we can
			   interpolate the slope in the gap between samples
			   Initialize where to inc past end of first frame
			*/
			PNMnextc();
			where = 0;
			while ((!PNMtimedout) && (PNMnextw() != 1)) {
				register uint32 change = PNMw;
				register int left = 0;

				if (change & 1) {
					/* This is a span record */
					int skip = 0;
					where += ((change + change) & ~3);

					/* Skip complete frames */
					while (where >= xyc) {
						now += wf;
						++skip;
						where -= xyc;
					}

					if (skip > 0) printf("now = %gs\n", now/1000000000.0);
				} else {
					/* Correct sample signedness */
					for (i=0; i<4; ++i) {
						switch ((where+i) % 6) {
						case 0:
						case 2:
							*(((uint8 *) &PNMw) + i) ^= 128;
						}
					}

					/* Add contribution from previous sample to sum */
					if (now > minstart) {
						double refd[4], newd[4];

						/* Map the old and new values */
						for (i=0; i<4; ++i) {
							refd[i] = ref[where+i];
							newd[i] = *(((uint8 *) &PNMw) + i);
						}

						FOR_RENDER {
							if ((MORE > 0) && (now > START)) {
								register double f = from[where];
								register double g = now - (wf - wt); /* start of gap between samples */
								register double fstart = ((START > f) ? START : f);
								register double fstop = ((g > STOP) ? STOP : g);
								register double gstart = ((START > g) ? START : g);
								register double gstop = ((now > STOP) ? STOP : now);
		
								/* Newly rendering this frame? */
								if (IMAGE == ((double *) 0)) {
									if (!(IMAGE = calloc(rendersp, xyc * sizeof(double)))) {
										ERROR(ERROR_EXPOSE,
										      "cannot allocate memory for TIK virtual exposure");
									}
								}
		
								/* Add portion of previous sample */
								if (fstart < fstop) {
									register double w = fstop - fstart;

									for (i=0; i<4; ++i) {
										SUM(where+i) += refd[i] * w;
									}

									/* Add-in fraction from real data frames */
									QUAL += (w / (now - f));
								}
		
								/* Add portion of gap as a slope */
								if (gstart < gstop) {
									register double wstart = (gstart - g) / (now - g);
									register double wstop = (gstop - g) / (now - g);
									register double wgap = gstop - gstart;

									for (i=0; i<4; ++i) {
										/* Compute gap end values, then average * wgap to get area */
										register double vstart = (refd[i] * (1.0 - wstart)) + (newd[i] * wstart);
										register double vstop = (refd[i] * (1.0 - wstop)) + (newd[i] * wstop);
										SUM(off+i) += (wgap * ((vstart + vstop) / 2.0));
									}
		
									/* Add-in gap fraction as if it is from one frame */
									QUAL += (wgap / (now - f));
								}
		
								/* Update more */
								if (now > STOP+wf) {
									/* Add-in data for other pixels */
									register int pos;

									for (pos=0; pos<xyc; ++pos) {
										register double f = from[pos];
										register double w = STOP - ((f < START) ? START : f);
										if (w < 0) w = 0;
										SUM(pos) = (SUM(pos) + (ref[pos] * w)) * DIVBY;
										++QUAL;
									}

									/* Correct exposure quality */
									QUAL /= (xyc / 4);

									/* Done with this virtual exposure; write the file */
									TIKrenderWriteFile(r);
								}
							}

							/* Still any images left incomplete? */
							left += (MORE > 0);
						}

						/* Check to see if any frames not yet rendered */
						if (left < 1) goto noneleft;
					}

					/* Update from pixel data */
					for (i=0; i<4; ++i) {
						ref[where+i] = *(((uint8 *) &PNMw) + i);
						from[where+i] = now;
					}
					where += 4;
					while (where >= xyc) {
						now += wf;
						where -= xyc;
					}
				}
			}

			/* Update image with last samples...
			   which we'll assume persist until stop,
			   because we can't know otherwise  ;-)
			   Also divide by time to get average values.
			*/
			FOR_RENDER
			if (MORE > 0) {
				/* Newly rendering this frame? */
				if (IMAGE == ((double *) 0)) {
					if (!(IMAGE = calloc(rendersp, xyc * sizeof(double)))) {
						ERROR(ERROR_EXPOSE,
						      "cannot allocate memory for UYVYYY virtual exposure");
					}
				}

				/* Update quality */
				++QUAL;

				for (where=0; where<xyc; ++where) {
					register double f = now - wf;
					register double w = STOP - ((f < START) ? START : f);

					if (w < 0) w = 0;
					SUM(where) = (SUM(where) + (ref[where] * w)) * DIVBY;
				}

				/* Done with this virtual exposure; write the file */
				TIKrenderWriteFile(r);
			}
			break;

		case wavePNM:
			/* Should be a P6 header next...
			   we know how to read that!
			*/
			p = P6reader();
			while (p) {
				register double f = now - wf;
				register double g = now - (wf - wt); /* start of gap between samples */
				register int left = 0;

				FOR_RENDER {
					if ((MORE > 0) && (now > START)) {
						register double fstart = ((START > f) ? START : f);
						register double fstop = ((g > STOP) ? STOP : g);
						register double gstart = ((START > g) ? START : g);
						register double gstop = ((now > STOP) ? STOP : now);
						register double wstart = (gstart - g) / (now - g);
						register double wstop = (gstop - g) / (now - g);
						register double wgap = gstop - gstart;

						/* Newly rendering this frame? */
						if (IMAGE == ((double *) 0)) {
							if (!(IMAGE = calloc(rendersp, xyc * sizeof(double)))) {
								ERROR(ERROR_EXPOSE,
								      "cannot allocate memory for PNM virtual exposure");
							}
						}

						/* Update quality */
						++QUAL;

						for (where=0; where<xyc; ++where) {
							/* Add portion of previous sample */
							if (fstart < fstop) {
								register double w = fstop - fstart;
								SUM(where) += (gamma[ref[where]] * w);
							}

							/* Add portion of gap as a slope */
							if (gstart < gstop) {
								/* Compute gap end values, then average * wgap to get area */
								register double vstart = (ref[where] * (1.0 - wstart)) + (gamma[p[where]] * wstart);
								register double vstop = (ref[where] * (1.0 - wstop)) + (gamma[p[where]] * wstop);
								SUM(where) += (wgap * ((vstart + vstop) / 2.0));
							}
						}

						/* Update more and write file */
						if (now > STOP+wf) {
							for (where=0; where<xyc; ++where) {
								SUM(where) = (SUM(where) * DIVBY);
								if (e_gamma != 1) {
									SUM(where) = pow(SUM(where), invgamma);
								}
							}

							/* Done with this virtual exposure; write the file */
							TIKrenderWriteFile(r);
						}
					}

					/* Still any images left incomplete? */
					left += (MORE > 0);
				}

				/* Check to see if any frames not yet rendered */
				if (left < 1) goto noneleft;

				/* Next image */
				now += wf;
				free(ref);
				ref = p;
				p = P6reader();
			}

			/* Did we get another? If so, discard it. */
			if (p) free(p);

			/* Update image with last samples...
			   which we'll assume persist until stop,
			   because we can't know otherwise  ;-)
			   Also divide by time to get average values.
			*/
			FOR_RENDER
			if (MORE > 0) {
				/* Newly rendering this frame? */
				if (IMAGE == ((double *) 0)) {
					if (!(IMAGE = calloc(rendersp, xyc * sizeof(double)))) {
						ERROR(ERROR_EXPOSE,
						      "cannot allocate memory for PNM virtual exposure");
					}
				}

				/* Update quality */
				++QUAL;

				for (where=0; where<xyc; ++where) {
					register double f = now - wf;
					register double w = STOP - ((f < START) ? START : f);

					if (w < 0) w = 0;
					SUM(where) = (SUM(where) + (gamma[ref[where]] * w)) * DIVBY;
					if (e_gamma != 1) {
						SUM(where) = pow(SUM(where), invgamma);
					}
				}

				/* Done with this virtual exposure; write the file */
				TIKrenderWriteFile(r);
			}

			break;

		default:
			goto badformat;
		}

		/* Clean-up */
noneleft:
		free(ref);
		free(from);

		/* Close the input file */
		if (dottik) {
			close(PNMfd);
		} else {
			stop_ffmpeg();
		}
	}
}

void
TIKrenderFile(int d, char *filename)
{
	dottik = d;
	infile = filename;
	rendersp = 0;
}

void
TIKrender(double start, double stop)
{
	/* Normalize start, stop and convert to nanoseconds */
	start *= 1000000000.0;
	stop *= 1000000000.0;
	if (stop < start) { double t = start; start = stop; stop = t; }

	/* When does the first frame start? */
	if ((rendersp == 0) || (start < minstart)) minstart = start;

	/* Queue it up */
	render[rendersp].start = start;
	render[rendersp].stop = stop;
	render[rendersp].qual = 0;
	render[rendersp].divby = 1.0 / (stop - start);
	render[rendersp].image = ((double *) 0);
	++rendersp;
}
