/*	tik.h

	Temporal Image Kentucky/Kontainer

	By Henry Dietz, June 2016

	"P%c\n"			// 6 or 5
	"# TIK V %d %s\n"	// 20160705 RGB or UYVYYY
	"# TIK X %d\n"		// X dim in pixels
	"# TIK Y %d\n"		// Y dim in pixels
	"# TIK Z %d\n"		// Maximum value
	"# TIK B %d\n"		// Begin time in nanoseconds (64-bit)
	"# TIK F %d\n"		// Frame time in nanoseconds
	"# TIK T %d\n"		// Shutter time in nanoseconds
	"# TIK R %d\n"		// Rolling shutter time in nanoseconds or 0
	"# TIK G %d\n"		// Gamma * 1000000
	// other comments, white space
	"%d %d\n"		// X, Y in samples
	"%d\n"			// Maximum value (same as Z)
	// RGB bytes for initial frame
	"\000"			// to ensure it doesn't look like "P"
	// Update records, (skip R G B)* or (skip U Y V Y Y Y)
	// where skip is 7 bits/byte, LSB first, 0x80 continued

	Note that a single P6 or P5 file can contain a sequence of
	images, but that shouldn't be confused with this format
	because the update records don't look like "P6" or "P5".
*/
		

#define	FFMPEG	"/usr/bin/ffmpeg"

/*	leave defined to make error models in log domain */
#define	LOG_ERR	1
#undef	LOG_ERR

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <jpeglib.h>
#include <math.h>
#include <time.h>

typedef	unsigned long long int	uint64;
typedef	long long int		int64;
typedef	unsigned int		uint32;
typedef	unsigned char		uint8;
typedef	signed char		int8;

extern double log(double x);
extern double pow(double x, double y);

typedef struct {
	double	sum[3];	/* Sum of pixel values merged */
	double	count;	/* How many were summed? */
	off_t	lseek;	/* Where was last output copy of this pixel? */
} tdciavg_t;


/*	Output message stuff... */
#define	INFO(...) do { if (e_interactive) { myinfo(); fprintf(stderr, __VA_ARGS__); } } while (0)
#define	UPDATE(...) do { if (e_interactive && (time(0) >= e_update + e_interactive - 1)) { myinfo(); fprintf(stderr, __VA_ARGS__); e_update = time(0); } } while (0)
#define	WARN(...) do { mywarn(); fprintf(stderr, __VA_ARGS__); } while (0)
#define	ERROR(EXIT, ...) do { myerror(); fprintf(stderr, __VA_ARGS__); exit(EXIT); } while (0)

typedef enum {
	ERROR_OK,
	ERROR_USAGE,
	ERROR_ARG,
	ERROR_READ,
	ERROR_OPEN,
	ERROR_CREATE,
	ERROR_FORMAT,
	ERROR_EXPOSE,
	ERROR_WRITE,
	ERROR_FIFO
} ERROR_t;

#define	E_N_MAX		1000000000

#define	WHAT_UNSPEC	0	/* Don't know what to do with filename */
#define	WHAT_ERRMOD	1	/* Make error model from video */
#define	WHAT_VIDEO	2	/* Encode video into .tik */
#define	WHAT_TIK	4	/* Extract images from a .tik */

#define	MAKE_UNSPEC	0
#define	MAKE_JPEG	1	/* Make stills as JPEG files */
#define	MAKE_PPM	2	/* Make stills as P6 PPM files */

extern	char	*myname;

extern	FILE	*TDCIfile;

extern	double	e_begin;
extern	char	*e_noisefile;
extern	double	e_fps;
extern	double	e_gamma;
extern	int	e_n;
extern	char	*e_outfile;
extern	char	*e_outtik;
extern	char	*e_outimage;
extern	double	e_percent;
extern	double	e_quality;
extern	double	e_t;
extern	int	e_what;
extern	int	e_interactive;
extern	time_t	e_update;
extern	int	e_maketype;


#include "color.h"
#include "reader.h"
#include "writer.h"
#include "noise.h"
#include "render.h"
#include "encode.h"
