/*	writer.c
*/

#include "tik.h"

/*	Output message stuff... */
void myinfo(void) { fprintf(stderr, "%s: ", myname); }
void mywarn(void) { fprintf(stderr, "%s: warning: ", myname); }
void myerror(void) { fprintf(stderr, "%s: error: ", myname); }

char *
tdci2jpeg(char *filename, double *image)
{
	/* Write image to filename as a JEPG
	   adjusts gamma as requested
	*/
	extern double pow(double x, double y);
	register FILE *outfile = fopen(filename, "wb");
	register uint8 *imout;
	register int i, j;
	register int wx = ((wavetype == waveUYVYYY) ? (wavex * 4) : wavex);
	register int wavexy = wx * wavey;
	register int wavexy3 = 3 * wavexy;

	if (!outfile) {
		return("could not be created\n");
	}

	/* Convert to 8 bits per pixel per color channel */
	imout = malloc(wavexy3);
	j = 0;
	if ((wavetype == waveUYVYYYold) || (wavetype == waveUYVYYY)) {
		for (i=0; i<wavexy3; i+=12) {
			/* CHDK live view data is:
			 U Y1 V Y2 Y3 Y4
			*/
			register double u = image[j+0] - 128.0;
			register double v = image[j+2] - 128.0;
			imout[i+0] = chdk_yuv2r(image[j+1], u, v);
			imout[i+1] = chdk_yuv2g(image[j+1], u, v);
			imout[i+2] = chdk_yuv2b(image[j+1], u, v);
			imout[i+3] = chdk_yuv2r(image[j+3], u, v);
			imout[i+4] = chdk_yuv2g(image[j+3], u, v);
			imout[i+5] = chdk_yuv2b(image[j+3], u, v);
			imout[i+6] = chdk_yuv2r(image[j+4], u, v);
			imout[i+7] = chdk_yuv2g(image[j+4], u, v);
			imout[i+8] = chdk_yuv2b(image[j+4], u, v);
			imout[i+9] = chdk_yuv2r(image[j+5], u, v);
			imout[i+10] = chdk_yuv2g(image[j+5], u, v);
			imout[i+11] = chdk_yuv2b(image[j+5], u, v);
			j += 6;
		}
	} else {
		/* Just convert RGB data from double into uint8 */
		for (i=0; i<wavexy3; ++i) {
			imout[i] = ((uint8) (image[i] + 0.5));
		}
	}

	if (e_maketype == MAKE_PPM) {
		/* Write PPM file */
		register int max = 0;

		/* Find actual max byte value for P6 header... */
		for (i=0; (i<wavexy3)&&(max<255); ++i) {
			if (imout[i] > max) max = imout[i];
		}

		/* Write P6 header */
		fprintf(outfile,
"P6\n"
"# Generated by %s %u\n"
"%d %d\n"
"%d\n",
			myname,
			VERSION,
			wx,
			wavey,
			max);

		/* Write the image data */
		fwrite(imout, wavexy3, sizeof(uint8), outfile);
	} else {
		/* Write JPEG file */
		struct jpeg_compress_struct cinfo;
		struct jpeg_error_mgr jerr;
		JSAMPROW row_pointer;

		cinfo.err = jpeg_std_error(&jerr);
		jpeg_create_compress(&cinfo);
		jpeg_stdio_dest(&cinfo, outfile);
		cinfo.image_width = wx;
		cinfo.image_height = wavey;
		cinfo.input_components = 3;
		cinfo.in_color_space = JCS_RGB;
		jpeg_set_defaults(&cinfo);
		jpeg_set_quality(&cinfo, ((int) (e_quality + 0.5)), 1);
		jpeg_start_compress(&cinfo, 1);

		/* Write it */
		while (cinfo.next_scanline < cinfo.image_height) {
			row_pointer = (JSAMPROW) (imout + (cinfo.next_scanline * 3 * wx));
			jpeg_write_scanlines(&cinfo, &row_pointer, 1);
		}

		jpeg_finish_compress(&cinfo);
	}

	/* Cleanup */
	free(imout);
	fclose(outfile);
	return(0);
}

